
(in-package "SB-C")

(declaim (ftype (sfunction (ctran ctran (or lvar null) t) (values))
                ir1-convert))

(defmacro combination-hook (&body body)
  (declare (ignore body))
  (compiler-error "illegal function call"))

(macrolet (;; Bind *COMPILER-ERROR-BAILOUT* to a function that throws
           ;; out of the body and converts a condition signalling form
           ;; instead. The source form is converted to a string since it
           ;; may contain arbitrary non-externalizable objects.
           (ir1-error-bailout ((start next result form) &body body)
             (with-unique-names (skip condition)
               `(block ,skip
                 (let ((,condition (catch 'ir1-error-abort
                                     (let ((*compiler-error-bailout*
                                            (lambda (&optional e)
                                              (throw 'ir1-error-abort e))))
                                       ,@body
                                       (return-from ,skip nil)))))
                   (ir1-convert ,start ,next ,result
                                (make-compiler-error-form ,condition ,form)))))))

  ;; Translate FORM into IR1. The code is inserted as the NEXT of the
  ;; CTRAN START. RESULT is the LVAR which receives the value of the
  ;; FORM to be translated. The translators call this function
  ;; recursively to translate their subnodes.
  ;;
  ;; As a special hack to make life easier in the compiler, a LEAF
  ;; IR1-converts into a reference to that LEAF structure. This allows
  ;; the creation using backquote of forms that contain leaf
  ;; references, without having to introduce dummy names into the
  ;; namespace.
  (defun ir1-convert (start next result form)
    (ir1-error-bailout (start next result form)
      (let ((*current-path* (or (gethash form *source-paths*)
                                (cons form *current-path*))))
        (cond ((step-form-p form)
               (ir1-convert-step start next result form))
              ((atom form)
               (cond ((and (symbolp form) (not (keywordp form)))
                      (ir1-convert-var start next result form))
                     ((leaf-p form)
                      (reference-leaf start next result form))
                     (t
                      (reference-constant start next result form))))
              (t
               (let ((opname (car form)))
                 (cond ((or (symbolp opname) (leaf-p opname))
                        (let ((lexical-def (if (leaf-p opname)
                                               opname
                                               (lexenv-find opname funs))))
                          (typecase lexical-def
                            (null
                             (ir1-convert-global-functoid start next result
                                                          form))
                            (functional
                             (ir1-convert-local-combination start next result
                                                            form
                                                            lexical-def))
                            (global-var
                             (ir1-convert-srctran start next result
                                                  lexical-def form))
                            (t
                             (aver (and (consp lexical-def)
                                        (eq (car lexical-def) 'macro)))
                             (ir1-convert start next result
                                          (careful-expand-macro (cdr lexical-def)
                                                                form))))))
                       ((or (atom opname) (not (eq (car opname) 'lambda)))
                        (ir1-convert start next result
				     (cons 'combination-hook form)))
                       (t
                        ;; implicitly (LAMBDA ..) because the LAMBDA
                        ;; expression is the CAR of an executed form
                        (ir1-convert-combination start next result
                                                 form
                                                 (ir1-convert-lambda
                                                  opname
                                                  :debug-name (debug-name
                                                               'lambda-car
                                                               opname))))))))))
    (values))
)
